/*
 ============================================================================
 Name        : HTTP.c
 Author      : Luis Tarrataca
 Version     :
 Copyright   : Your copyright notice
 Description : HowTo HTTP Requests Posix Style
 ============================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>

#define TRUE 1
#define FALSE 0

#define MAXIMUM_NUMBER_OF_DIGITS 10
#define MAXIMUM_NUMBER_OF_BYTES_TO_READ 500
#define MAXIMUM_NUMBER_OF_BYTES_OF_HTML_REQUEST_HEADER 200
#define MAXIMUM_NUMBER_OF_BYTES_OF_HTML_RESPONSE_HEADER 500
#define MAXIMUM_NUMBER_OF_BYTES_OF_HTML_RESPONSE_BODY 10000

struct sockaddr* getDestinationSocket( char* url ){

	printf("[[ getDestinationSocket ]]\n");

	char* index = (char*) memchr( url, '/', strlen( url ) );
	int numberOfCharactersOfServerURL = index - url;
	char* serverURL = (char*) calloc( numberOfCharactersOfServerURL + 1, sizeof( char ) );
	memcpy( serverURL, url, numberOfCharactersOfServerURL );
	*( serverURL + numberOfCharactersOfServerURL + sizeof( char ) ) = '\0';

	struct addrinfo* urlInformation    = NULL;
	struct addrinfo* auxiliarPointer   = NULL;
	struct sockaddr* destinationSocket = NULL;
	struct addrinfo hints;

	memset( &hints, 0, sizeof( hints ));
	hints.ai_family = AF_INET; // We only want IPv4 addresses
	hints.ai_protocol = SOCK_STREAM; // TCP FTW =)

	int error;


	printf("\tResolving URL: %s\n", serverURL );

	/* resolve the domain name into a list of addresses */
	error = getaddrinfo( serverURL, NULL, &hints, &urlInformation );
	if (error != 0)
	{
		// Convert to a human readable error
		fprintf(stderr, "\tError in getaddrinfo: %s\n", gai_strerror(error));
		return NULL;
	}

	/* loop over all returned results and do inverse lookup */
	for( auxiliarPointer = urlInformation; auxiliarPointer != NULL; auxiliarPointer = auxiliarPointer->ai_next )
	{
		// This is a constant defined in the library netdb.h
		char hostname[ NI_MAXHOST ] = "";

		error = getnameinfo( auxiliarPointer->ai_addr, auxiliarPointer->ai_addrlen, hostname, NI_MAXHOST, NULL, 0, 0 );
		if( error != 0 )
		{
			fprintf(stderr, "\t[ERROR] in getnameinfo: %s\n", gai_strerror(error));
			continue;
		}
		if( *hostname != '\0' ){
			printf("\tIP Address: %s\n", hostname);

			if( destinationSocket == NULL ){ destinationSocket = auxiliarPointer->ai_addr; }
		}

	}

	return destinationSocket;

}

int connectToDestinationSocket( struct sockaddr* destinationSocket ){

	int socketFd = socket( AF_INET, SOCK_STREAM, 0 );

	((struct sockaddr_in*) destinationSocket)->sin_port = htons( 80 ); //HTTP Port

	printf("[[ connectToDestinationSocket ]]\n");

	int result = connect( socketFd, destinationSocket, sizeof( struct sockaddr_in ) );

	if( result != 0 ){
		printf("\t[ERROR] :: Unable to connect with destination!\n" );
	}else{
		printf("\tConnected with destination!\n" );
	}

	return socketFd;

}

int connectToDestinationSocketViaIPAddress( char* ipAddress, char* port ){

	struct sockaddr_in dest;

   	int socketFd = socket(AF_INET, SOCK_STREAM, 0);

   	memset( &dest, 0, sizeof( dest ) );                /* zero the struct */
   	dest.sin_family = AF_INET;
   	dest.sin_addr.s_addr = htonl(inet_addr( ipAddress )); /* set destination IP number - localhost, 127.0.0.1*/
   	dest.sin_port = htons( atoi( port ) );           /* set destination port number */

	printf("[[ connectToDestinationSocket ]]\n");

	int result = connect( socketFd, (struct sockaddr *) &dest, sizeof( struct sockaddr_in ) );

	if( result != 0 ){
		printf("\t[ERROR] :: Unable to connect with destination!\n" );
	}else{
		printf("\tConnected with destination!\n" );
	}

	return socketFd;


}

void sendHTMLHeaderRequest( int socketFd, char* url, int printHTMLHeaderRequestFlag ){

	char* HTMLRequestHeader  = (char*) calloc( MAXIMUM_NUMBER_OF_BYTES_OF_HTML_REQUEST_HEADER , sizeof( char ) );
	strcat( HTMLRequestHeader, "GET http://" );
	strcat( HTMLRequestHeader, url );
	strcat( HTMLRequestHeader, " HTTP/1.0\r\n");
	strcat( HTMLRequestHeader, "Accept: text/plain, text/html, text/*\r\n");
	strcat( HTMLRequestHeader, "\r\n" );

	int HTMLRequestHeaderNumberOfBytes = strlen( HTMLRequestHeader ) * sizeof( char );

	printf("[[ sendHTMLHeaderRequest ]]\n");
	printf("\tSending HTML request message with %d bytes\n", HTMLRequestHeaderNumberOfBytes );
	send( socketFd, HTMLRequestHeader, HTMLRequestHeaderNumberOfBytes, 0 );

	if( printHTMLHeaderRequestFlag != 0 ){

		printf("\t-----HTML Request Header-----\n");
		printf("%s", HTMLRequestHeader );

	}

	return;

}


char* getHTMLHeaderResponse( int socketFd, int printHTMLHeaderResponseFlag ){

	char* HTMLResponseHeader = (char*) calloc( MAXIMUM_NUMBER_OF_BYTES_OF_HTML_RESPONSE_HEADER, sizeof( char ) );
	char characterRead;

	int HTMLResponderHeaderNumberOfBytesToRead = 1 * sizeof( char );
	int numberOfBytesRead = 0;
	int totalNumberOfBytesRead = 0;

	printf("[[ getHTMLHeaderResponse ]]\n");

	printf( "\tReading HTML response header...\n" );

	while( TRUE ){

		numberOfBytesRead = recv( socketFd, &characterRead, HTMLResponderHeaderNumberOfBytesToRead, 0 );
		*( HTMLResponseHeader + totalNumberOfBytesRead ) = characterRead;
		totalNumberOfBytesRead += numberOfBytesRead;

		// According to RFC 2616 body starts after a CR LF CR LF character sequence
		if( strlen( HTMLResponseHeader ) > 4 &&
				HTMLResponseHeader[ strlen( HTMLResponseHeader ) - 1 ] == '\n' &&
				HTMLResponseHeader[ strlen( HTMLResponseHeader ) - 2 ] == '\r' &&
				HTMLResponseHeader[ strlen( HTMLResponseHeader ) - 3 ] == '\n' &&
				HTMLResponseHeader[ strlen( HTMLResponseHeader ) - 4 ] == '\r' ){

			break;
		}
	}

	*( HTMLResponseHeader + totalNumberOfBytesRead ) = '\0';

	printf("\tRead %d bytes\n", totalNumberOfBytesRead );

	if( printHTMLHeaderResponseFlag != 0 ){

		printf("\t-----HTML Response Header-----\n");
		printf("%s", HTMLResponseHeader );

	}

	return HTMLResponseHeader;

}

int getContentLength( char* HTMLHeaderResponse ){

	printf( "[[ getContentLength ]]\n" );

	char* stringToSearch = "Content-Length:";
	char* index = strstr( HTMLHeaderResponse, stringToSearch );

	if( index == NULL ){

		printf("\tWarning :: HTML Response Header does not contain a header!\n");
		return -1;
	}

	index = ( strlen( stringToSearch ) +  1 ) * sizeof( char ) + index;

	char* contentLengthString = ( char* ) calloc( MAXIMUM_NUMBER_OF_DIGITS, sizeof(char) );

	int counter = 0;

	while( *( index + counter ) != '\r' ){

		*( contentLengthString +  counter ) = *( index + counter );

		counter++;

	}

	int contentLength = atoi( contentLengthString );
	printf("\tContent-Length:%d\n", contentLength );

	return contentLength;

}

void getHTMLMessageBodyV1( int socketFd, int contentLength ){

	printf("[[ getHTMLMessageBodyV1 ]]\n");

	int byteCounter = 0;


	int numberOfBytesToTryToRead = MAXIMUM_NUMBER_OF_BYTES_TO_READ * sizeof( char );
	char* buffer = (char*) malloc( numberOfBytesToTryToRead );

	while( TRUE ){

		memset( buffer, 0, numberOfBytesToTryToRead );
		byteCounter += recv( socketFd, buffer, numberOfBytesToTryToRead, 0 );
		printf("%s", buffer );

		if( byteCounter  == contentLength ){ break; }

	}

	printf("\n\tReceived %d bytes out of %d bytes\n", byteCounter, contentLength );

	return;
}


void getHTMLMessageBodyV2( int socketFd ){

	printf("[[ getHTMLMessageBodyV2 ]]\n");

	int numberOfBytesRead = 0;
	int byteCounter = 0;

	int numberOfBytesToTryToRead = MAXIMUM_NUMBER_OF_BYTES_TO_READ * sizeof( char );
	char* buffer = (char*) malloc( numberOfBytesToTryToRead );

	while( TRUE ){

		memset( buffer, 0, numberOfBytesToTryToRead );
		numberOfBytesRead = recv( socketFd, buffer, numberOfBytesToTryToRead, 0 );
		byteCounter += numberOfBytesRead;
		printf("%s", buffer );

		if( numberOfBytesRead  < numberOfBytesToTryToRead ){ break; }

	}

	printf("\n\tReceived %d bytes\n", byteCounter );

	return;
}

void connectExampleViaURL(){

	char* url = "sistemasoperacionais.com/index.html";
	//char* url = "web.ist.utl.pt/luis.tarrataca/hello.html";
	//char* url = "www.cefet-rj.br/";

	struct sockaddr* destinationSocket = getDestinationSocket( url );
	int socketFd = connectToDestinationSocket( destinationSocket );
	sendHTMLHeaderRequest( socketFd, url, FALSE );
	char* HTMLResponseHeader = getHTMLHeaderResponse( socketFd, TRUE );
	getHTMLMessageBodyV2( socketFd );

	return;
}

void connectExampleViaIPAddress(){

	char* ipAddress = "127.0.0.1";
	char* port = "80";

	int socketFd = connectToDestinationSocketViaIPAddress( ipAddress, port );
	sendHTMLHeaderRequest( socketFd, ipAddress, FALSE );
	char* HTMLResponseHeader = getHTMLHeaderResponse( socketFd, TRUE );
	getHTMLMessageBodyV2( socketFd );

	return;
}

int main( int argc, char** argv ){

	connectExampleViaURL();
}